/*
 * Copyright (C) 1995  Sam Moore, Warren Jasper
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/*
 * a2dc.h
 */

/* Default device major number */

/* #define ADC_BASE     0x300  */ /* Default base address */
/* #define ADC_IRQ         10  */ /* Default irq          */


#ifndef DEFAULT_MAJOR_DEV
#define DEFAULT_MAJOR_DEV  31    /* Default Major Device Number */
#endif
#define DEFAULT_FREQ       100   /* Default frequency           */
#define AD_CHANNELS         16   /* 16 Single Ended or 8 Diff.  */
#define DIO_CHANNEL         16   /* dio minor channel number    */

#define PORT_MIN         0x200   /* minimum base address used in autodetecting */
#define PORT_MAX         0x3e0   /* maximum base address allowed by board      */
#define PORT_STEP        0x20    /* base step in enhanced mode (32 bytes)      */
#define PORT_SIZE        0x14    /* uses 20 bytes of registers                 */

/* I/O Register locations */

/* NOTE: base_reg to be found dynamically at boot time. Not hard coded into driver */

#define LSB_AND_CHNLS    (base_reg)    /* A/D Data & Channel Register      */
#define MSB_DATA_BYTE    (base_reg+1)    
#define MUX_SCAN_LIMITS  (base_reg+2)  /* Channel Mux Scan Limits Register */
#define DIO_REG          (base_reg+3)  /* 4 Bit Digital I/O Register       */
#define STATUS_REG       (base_reg+8)  /* Status Register                  */
#define CNTRL_REG        (base_reg+9)  /* DMA, Interrupt & Trigger Control */
#define PACER_CLOCK_REG  (base_reg+10) /* Pacer Clock Control Register     */
#define MISC_REG         (base_reg+11) /* Analog Input Range Reg., ETC     */

#define COUNTERA_0_DATA  (base_reg+12) /* Pacer Clock A Counter 0 Register */
#define COUNTERA_1_DATA  (base_reg+13) /* Pacer Clock A Counter 1 Register */
#define COUNTERA_2_DATA  (base_reg+14) /* Pacer Clock A Counter 2 Register */
#define COUNTERA_CONTROL (base_reg+15) /* Pacer Clock A Control  Register  */

#define COUNTERB_0_DATA  (base_reg+16) /* Pacer Clock B Counter 0 Register */
#define COUNTERB_1_DATA  (base_reg+17) /* Pacer Clock B Counter 1 Register */
#define COUNTERB_2_DATA  (base_reg+18) /* Pacer Clock B Counter 2 Register */
#define COUNTERB_CONTROL (base_reg+19) /* Pacer Clock B Control  Register  */

/* Errors */

#define BAD_CHANNEL        1
#define BAD_COUNT          2
#define BAD_GAIN           3
#define NO_ENHANCED_MODE   4
#define BAD_SPEED          5
#define ADC_BUSY           6
#define INT_ALREADY_ACTIVE 7

/* General defines */

#define TRUE  1
#define FALSE 0

/*
#define DEBUG 
*/


/* DT Connection is not supported */

#define DT_CONNECT FALSE

/* Interrupt context data. Controls interrupt service handler */

typedef struct ADC_REC {
  WORD base;           /* Base address of ADC board                */
  BYTE ir;             /* shifted bits for board 0-7, 10->0, 11->1 */
  int  irq;            /* IRQ inturrepts (0-7, 10, 11)             */
  LONG freq;           /* Complete frequency value                 */
  int  busy;           /* busy = TRUE, free = FALSE                */
  WORD ctr1;           /* MSB of estimated "freq"                  */
  WORD ctr2;           /* LSB of estimated "freq"                  */
  BYTE pacerReg;       /* value of pacer clock control register    */
  BYTE  num_chan;      /* 8 differential or 16 single ended        */
} BoardRec;

/* Values kept for each channel. */

typedef struct CHANNEL_REC {
  int  open;            /* Is channel device open()                 */
  int  adc_dio_pretrig; /* TRUE = write dio before sampling         */
  LONG count;           /* Number of samples requested              */
  BYTE dio_value;       /* DIO value to be written before sampling  */
  BYTE mode;            /* 0 =  Soft Trigger                        */
                        /* 2 = External Trigger                     */
                        /* 4 = Pacer Clock                          */
                        /* 6 = Pacer Clock External Trig Enable     */
  BYTE gain;            /* Voltage range                            */
  BYTE lowChan;         /* Starting Channel of MUX scan limit       */
  BYTE hiChan;          /* Ending Channel of MUX scan limit         */
} ChanRec;


/*************************************************************************
* STATUS_REG         base_reg+8             Status Register              *
**************************************************************************/
#define ADSR_INT        0x10      /* INT=1 External pulse received, INT=0 waiting */
#define ADSR_MUX        0x20      /* MUX=1 16 Single Ended, MUX=0 8 Differential  */
#define ADSR_UB         0x40      /* U/B=1 Unipolar mode, U/B=0 Bipolar mode      */
#define ADSR_EOC        0x80      /* EOC=1, A/D is busy, EOC=0 free               */

/*************************************************************************
* CNTRL_REG          base_reg+9        DMA, Interrupt & Trigger Control  *
**************************************************************************/
#define CNTL_TS0       0x1    /* TS0=0 External Trigger, TS0=1 Pacer Clock */
#define CNTL_TS1       0x2    /* TS1=0 Software triggered A/D only         */
#define CNTL_DMA       0x4    /* DMA Disabled = 0, DMA Enabled = 1         */
#define CNTL_INTE      0x80   /* Interrupts Disabled = 0, Enabled = 1      */

/*************************************************************************
* MISC_REG        base_reg+11         Range, and Exteded Features        *
**************************************************************************/

#define ENHANCED_BIT  (0x10)  /* Enhanced Mode Enabled = 1, Disabled = 0 */
#define MISC_OVERRUN  (0x20)  /* No overrun = 0, FIFP buffer full = 1    */
#define MISC_PRETRIG  (0x40)  /* Pretrigger Enabled = 1, Disable = 0     */
#define MISC_DT       (0x80)  /* DT-Connect Enable = 1, Disable = 0      */

/*************************************************************************
* Constants for dealing with the 8254 counters                           *
**************************************************************************/

#define MODE0 0x0
#define MODE1 0x2
#define MODE2 0x4
#define MODE3 0x6
#define MODE4 0x8
#define MODE5 0xa

#define C0 0x00
#define C1 0x40
#define C2 0x80

#define LATCH    0x00
#define LSBONLY  0x10
#define MSBONLY  0x20
#define LSBFIRST 0x30

#define S0       0x00
#define S1       0x02

#define PACKETSIZE  512
#define MAX_COUNT   16384
